(** adapted from Berke Durak's fft.ml *)

module Complex =
  struct
	type t = float * float
	let i = (0.0,1.0)
	let norm_squared (x,y) = (x *. x) +. (y *. y)
	let norm z = sqrt (norm_squared z)
	let from_polar rho theta = (rho *. (cos theta),rho *. (sin theta))
	let conjugate (x1,y1) = (x1,-.y1)
	let add (x1,y1) (x2,y2) = (x1 +. x2, y1 +. y2)
	let sub (x1,y1) (x2,y2) = (x1 -. x2, y1 -. y2)
	let mul (x1,y1) (x2,y2) = (x1 *. x2 -. y1 *. y2,x1 *. y2 +. y1 *. x2)
	let div z1 z2 =
	  let f (x,y) = x *. x -. y *. y in
	  let z2' = conjugate z2 in
	  let (x,y) = mul z1 z2 in
	  let t = f z2 in
	  (x /. t, y /. t)
  end

module FFT =
  struct
    module C = Complex
    open Complex    
    let odd x = x land 1 = 1
    let even x = x land 1 = 0
    
    let rec lg2 = function
    	0 -> 0
      |	x -> 1 + (lg2 (x / 2))
    		
    let rec is_power_of_two k =
      k = 0 or (even k && is_power_of_two (k lsr 1))
    	
    	let bit_reversal_table m =
    	  let n = 1 lsl m in
    	  let t = Array.make n 0 in
    	  for i = 1 to n - 1 do
    		if odd i then
    		  t.(i) <- (t.(i lsr 1) lsr 1) lor (n lsr 1)
    		else
    		  t.(i) <- t.(i lsr 1) lsr 1
    	  done;
    	  t
    		
    	let shuffle t a b =
    	  let n = Array.length a in
    	  for i = 0 to n - 1 do
    		b.(i) <- a.(t.(i))
    	  done
    		
    	let dft lgn dir a =
    	  let n = 1 lsl lgn in
    	  let pi2 = 8.0 *. (atan 1.0) in
    	  for s = 1 to lgn do
    		let m = 1 lsl s in
    		let w_m = C.from_polar 1.0 (dir *. pi2 /. (float_of_int m)) in
    		let rec loop_1 w j =
    		  if j < m / 2 then
    			begin
    			  let rec loop_2 k =
    				if k < n then
    				  begin
    					let k' = k + m / 2 in
    					let t = C.mul w a.(k') in
    					let u = a.(k) in
    					a.(k) <- add t u;
    					a.(k') <- sub t u;
    					loop_2 (k + m)
    				  end
    				else
    				  loop_1 (C.mul w w_m) (j +1)
    			  in
    			  loop_2 j
    			end
    		  else
    			()
    		in
    		loop_1 (1.0,0.0) 0
    	  done
    		
    	let fft a =
    	  let n = Array.length a in
    	  let b = Array.make n (0.0,0.0) in
    	  let lgn = lg2 (n - 1) in
    	  let t = bit_reversal_table lgn in
    	  let b = Array.make n (0.0,0.0) in
    	  shuffle t a b;
    	  dft lgn (1.0) b;
    	  b
    		
    	let prepare_fft order dir =
    	  let b = Array.make order (0.0,0.0)
    	  in
    	  let lgn = lg2 (order - 1) in
	  let t = bit_reversal_table lgn
	  in
	  (fun a -> 
		shuffle t a b;
		dft lgn 1.0 b)
end

