(*
   Useful notes:
   - http://www.eptools.com/tn/T0001/INDEX.HTM
   - http://www.clarku.edu/~djoyce/complex/
   - http://en2.wikipedia.org/wiki/Cooley-Tukey_FFT_algorithm
*)

(* special multiplier that optimizes out multiplication by 0 1 and -1 *)
let lift x = <x>;;
let pi = 4.0 *. atan 1.0;;
let ( +..) ( (r,s)) ( (t,u)) =  (r+.t, s+.u);;
let ( *..) (r,s) (t,u) = ((r *. t) -. (s *. u), (r *. u) +. (s *. t))
let ( +...) a b = < (fun ( (r,s)) ( (t,u)) ->  (r+.t, s+.u)) ~a ~b >;;
let ( -...) a b = < (fun ( (r,s)) ( (t,u)) ->  (r-.t, s-.u)) ~a ~b >;;
let ( *...) a b = match b with
            (0.,0.) -> < (0., 0.) >
          | (1.,0.) -> < (fun (r,s) -> (r, 0.)) ~a >
          | (0.,1.) -> < (fun (r,s) -> (0., s)) ~a >
          | (1.,1.) -> a
          | (-1.,0.) -> < (fun (r,s) -> ((~-. r), 0.)) ~a >
          | (0.,-1.) -> < (fun (r,s) -> (0., (~-. s))) ~a >
          | (-1.,-1.) -> < (fun (r,s) -> (~-. r, ~-. s)) ~a >

          | (t, 0.) -> < (fun (r,s) ->  ((r *. t), (s *. t))) ~a >
          | (t, 1.) -> < (fun (r,s) ->  ((r *. t) -. s, r +. (s *. t))) ~a >
          | (0., u) -> < (fun (r,s) ->  ((s *. u), (r *. u))) ~a >
          | (1., u) -> < (fun (r,s) ->  (r -. (s *. u), (r *. u) +. s)) ~a >
          | (t, -1.) -> < (fun (r,s) ->  ((r *. t) -. (~-. s), (~-. r) +. (s *. t))) ~a >
          | (-1., u) -> < (fun (r,s) ->  ((~-. r) -. (s *. u), (r *. u) +. (~-. s))) ~a >

          | (t,u)   -> < (fun (r,s) ->  ((r *. t) -. (s *. u), (r *. u) +. (s *. t))) ~a >

(*
 *  Evaluates continuation k with an argument f' such that for f'(i) is
 *  a code fragment bound to the result of evaluating f(i) for all 0<=i<n
 *)


let omega j n = let z = 2. *. pi *. ((float j) /. (float n)) in (cos z , sin z)
let rec upto = function 0 -> [] | n -> (n-1)::upto (n-1)
let rec fold f = function (h::t) -> List.fold_right f t h | _ -> raise Not_found;;

let rec memoize n f k =
        if n .< 0
        then k f
        else  < let z = ~(f n) in ~ (memoize (n-1) (fun i -> if i=n then <z> else f i) k)  >

let rec fft radix length inputs k =
    let rec butterfly length inputs k slice =

        let rec continuation memoized_inputs =
          butterfly length
                    (fun i -> if i*radix/length == slice
                              then (memoized_inputs (i mod length/radix))
                              else inputs i)
                    k (slice-1)

        and memoized_continuation inputs =
          memoize (length/radix-1) inputs continuation

        in if (slice .< 0)
           then k inputs
           else if (length/radix .< radix)
                then memoized_continuation inputs
                else butterfly (length/radix)
                               (fun i -> inputs (i + length * slice / radix))
                               (fun inputs -> memoized_continuation (merge (length/radix) inputs))
                               (radix-1)
        
    and merge length inputs i =
      fold (+...) (List.map
                     (fun slice ->
                          (inputs (slice * (length/radix) + (i mod (length/radix))))
                       *...
                          ((omega (slice * (i mod (length/radix))) length)
                       *..
                           (omega (radix * slice * i / length) radix)))
                     (upto radix))

    in butterfly length inputs (fun inputs' -> k (merge length inputs'))


;;

let radix = 4;;
let inputsize = 16;;
let fa = < fun a0 a1 a2 a3 a4 a5 a6 a7 a8 a9 a10 a11 a12 a13 a14 a15 ->
           ~(
              fft radix inputsize
                 (fun x -> match x with
                        0->(<a0>)  |1->(<a1>)   |2->(<a2>)   |3->(<a3>)
                      |4->(<a4>)   |5->(<a5>)   |6->(<a6>)   |7->(<a7>)
                      |8->(<a8>)   |9->(<a9>)   |10->(<a10>) |11->(<a11>)
                      |12->(<a12>) |13->(<a13>) |14->(<a14>) |_-> (<a15>))
                 (fun x ->
                    < ~(x 0), ~(x 1), ~(x 2), ~(x 3),
                      ~(x 4), ~(x 5), ~(x 6), ~(x 7),
                      ~(x 8), ~(x 9), ~(x 10), ~(x 11),
                      ~(x 12), ~(x 13), ~(x 14), ~(x 15)
                    >)
                 (radix - 1)
            )
         >



(* naive staging *)
let rec power n =
  if n = 0 then < fun x -> 1 >
           else < fun x -> n * (~ (power (n-1)) x) >
;;

(* better staging *)
let rec power n =
       if n = 0        then < fun x -> 1 >
  else if n mod 2 != 0 then < fun x -> n * (~ (power (n-1)) x) >
                       else < fun x -> let y = (~ (power (n/2)) x) in y * y >
;;

let rec permute k n =
  if k = 1 then n
           else n * (permute (k-1) (n-1))
;;

let rec permute k =
  if k = 1 then < fun n -> n >
           else < fun n -> n *. ( ~(permute (k-1)) (n-.1.)) >
;;

let rec permute k =
       if k = 1        then < fun n -> n >
  else if k mod 2 = 0  then < fun n -> ( ~(permute (k/2)) n) *. ( ~(permute (k/2)) (n-. ~(lift(float(k/2))))) >
                       else < fun n -> n *. ( ~(permute (k-1)) (n-.1.)) >
;;



(* best *)
let lift x = < x >;;
let rec permute k r =
       if k = 1        then < fun n -> n -. ~(lift(float r)) >
  else if k mod 2 = 0  then
    < fun n ->
      ( ~(permute (k/2) (r+k/2)) n)
        *.
        ( ~(permute (k/2) r) n) >
  else < fun n -> n *. ( ~(permute (k-1) (r+1)) (n-. ~(lift(float r)))) >
;;






(*******************************************************************************)

(* good *)
let rec permute k =
  if k = 1 then < fun n -> n >
           else < fun n -> n * ( ~(permute (k-1)) (n-1)) >
;;

(* better; splits the iteration-space in half; log(n) recursions instead of O(n) *)
let lift x = < x >;;
let rec permute k =
       if k = 1        then < fun n -> n >
  else if k mod 2 = 0  then < fun n -> ( ~(permute (k/2)) n) * ( ~(permute (k/2)) (n - ~(lift (k/2)))) >
                       else < fun n -> n * ( ~(permute (k-1)) (n-1)) >
;;

(* best; also computes the multiplication argument at each stage by
 * subtracting a constant from the original instead of subtracting one
 * from the last multiplicative argument
 *)
let lift x = < x >;;
let rec permute k r =
       if k = 1        then < fun n -> n - ~(lift( r)) >
  else if k mod 2 = 0  then
    < fun n ->
      ( ~(permute (k/2) (r+k/2)) n)
        *
        ( ~(permute (k/2) r) n) >
  else < fun n -> n * ( ~(permute (k-1) (r+1)) (n- ~(lift( r)))) >
;;




let f0 = ! fa;;

let sample0 = fun i -> ( (float i, float i));;
f0 (sample0 0) (sample0 1) (sample0 2) (sample0 3)
(sample0 4) (sample0 5) (sample0 6) (sample0 7)
(sample0 8) (sample0 9) (sample0 10) (sample0 11)
(sample0 12) (sample0 13) (sample0 14) (sample0 15)
;;



