(*
   Useful notes:
   - http://www.eptools.com/tn/T0001/INDEX.HTM
   - http://www.clarku.edu/~djoyce/complex/
   - http://en2.wikipedia.org/wiki/Cooley-Tukey_FFT_algorithm
*)

let rec map f l = match l with [] -> [] | a::b -> (f a)::(map f b);;

(* complex math *)
let pi = 4.0 *. atan 1.0;;
type complex = Complex of (float * float);;
let lift (Complex(x,y)) = .<(Complex(x,y))>.;;
let ( +..) (Complex (r,s)) (Complex (t,u)) = Complex (r+.t, s+.u);;
let ( +...) a b = .< (fun (Complex (r,s)) (Complex (t,u)) -> Complex (r+.t, s+.u)) .~a .~b >.;;
let ( -..) (Complex (r,s)) (Complex (t,u)) = Complex (r-.t, s-.u);;
let ( -...) a b = .< (fun (Complex (r,s)) (Complex (t,u)) -> Complex (r-.t, s-.u)) .~a .~b >.;;
let ( *..) (Complex (r,s)) (Complex (t,u)) = Complex ((r *. t) -. (s *. u), (r *. u) +. (s *. t));;
let ( *...) a b = .< 
     (fun (Complex (r,s)) (Complex (t,u)) ->
          Complex ((r *. t) -. (s *. u), (r *. u) +. (s *. t))) .~a .~b >.
let omega j n =
  let z = 2.0 *. pi *. ((float j) /. (float n))
  in  Complex (cos z , sin z);;

(*
 *  Evaluates continuation k with an argument g such that for g(i) is
 *  a code fragment bound to the result of evaluating f(i) for all 0<=i<n
 *)
exception BindVarExn of (int * int);;
let rec bindvars    n f k = bindvars' (n-1) (fun idx -> raise (BindVarExn (idx, n))) f k
    and augment     g i v = fun i' -> if (i' == i) then v else g i'
    and bindvars' n g f k =
             if n < 0 then k g
                      else .< (fun z -> .~ (bindvars' (n-1) (augment g n .<z>.) f k)) (.~(f n)) >.;;
(*
     use this line instead for less bulky code printouts; however the memoization
     structure is not as visually obvious.

                      else .< let z = .~(f n) in .~ (bindvars' (n-1) (augment g n .<z>.) f k) >.
*)



(* a "known good" fft, although this rearranges and rotates some of its outputs (many FFTs do) *)
(*
#use "fft-baseline.ml";;
*)
(*
let known_good_fft n f =
  let rec populate a i = if i < 0 then a else (a.(i) <- (match (f i) with Complex(x,y) -> (x,y)); (populate a (i-1))) in
  let a = FFT.fft (populate (Array.create(n) (0.0, 0.0)) (n-1)) in
  fun i -> a.(i)
let sample_known_good x = Complex(float x,float x);;
let known_good = map (fun (x,y) -> (round x, round y)) (map (known_good_fft i sample0) (upto (i-1)));;
*)

(* sums the code outputs of a function for inputs 0..n *)
exception SumException;;
let rec sum n f = if n < 0 then raise SumException
                           else if n == 0 then f 0
                           else .< .~((f n) +... (sum (n-1) f)) >.;;

(*
 *  Staged, memoized fft.
 *
 *  radix   - the radix of the fft; can be any value as long as the length of the input is a power of the radix
 *  n2      - the length of the input divided by the radix (ie length=radix*n2)
 *  f       - a function mapping 0..(radix*n2-1) to code fragments representing the input values (int -> <Complex>)
 *  k       - a continuation; its argument has the same type as f and represents the transformed signal
 *)
let rec fft radix n2 f k =
        let rec coeff i slice            = (omega (slice * (i mod n2)) (radix*n2)) *.. (omega ((i / n2) * slice) radix)
            and contribution f' i slice  = .< .~((f' (slice * n2 + (i mod n2))) *... (lift (coeff i slice))) >.
            and merge f' i               = sum (radix-1) (contribution f' i)

            in if n2 == 0 then .< .~ (k f) >.
                          else subfft radix n2 (radix-1) f (fun f' -> .< .~ (k(merge f')) >.)

    (* performs an "in place" fft on values whichslice*n2..((whichslice+1)*n2-1), generating bindings in the process *)
    and subfft radix n2 whichslice f k = if (whichslice < 0) then .< .~ (k f) >. else
        let rec f' i = .< .~ (f (i + n2 * whichslice)) >.
            and n2'  = n2/radix

        in  fft radix n2' f'
            (fun f_sub -> bindvars n2 f_sub
                         (fun bound_f_sub -> subfft radix n2 (whichslice-1)
                                       (fun i -> if i/n2 == whichslice then (bound_f_sub (i mod n2)) else (f i))
                                       k));;

(* some tests *)
let rec upto z = match z with 0 -> 0::[] | n -> n::(upto (z-1));;

let liftf fl = .<fl>.;;
let sample x = .< Complex(.~(liftf(float x)), .~(liftf(float x))) >.;;
let round i = if i -. (floor i) > 0.5 then ceil i else floor i;;

let sample0 = fun i -> (Complex (float i, float i));;
let unrolled radix inputsize = .< fun a -> .~(fft radix (inputsize/radix) (fun x -> .<a x>.) (fun x -> .<x>.)) >.;;

let unrolled radix inputsize map i =
    .< .~ (fft radix (inputsize/radix) (fun x -> map x) (fun x -> .< .~(x i) >.)) >.;;
let mapper i = (Complex (float i, float i));;
let radix = 2;;
let inputsize = 4;;
(*
let fa = ( .< fun a b c d ->
      .~ (bindvars 4
               (fun i -> unrolled radix inputsize (fun i -> match i with 0 -> (.<a>.)|1->(.<b>.)|2->(.<c>.)|_->(.<d>.)) i)
               (fun g -> .< (.~(g 0), .~(g 1), .~(g 2), .~(g 3)) >. )
         ) >. );;
*)
let fa = .< fun a b c d ->
       .~ (
            let map = (fun i -> match i with 0 -> (.<a>.)|1->(.<b>.)|2->(.<c>.)|_->(.<d>.))
            in fft radix (inputsize/radix) (fun x -> map x) (fun x -> .< (.~(x 0),.~(x 1),.~(x 2),.~(x 3)) >.)
          ) >. ;;
let f0 = .! fa;;
f0 (sample0 0) (sample0 1) (sample0 2) (sample0 3);;
(*
let test = unrolled 2 4;;

let foo = fun a -> .< .~ test a a >.;; 
*)
