
(* For FFT, see http://www.eptools.com/tn/T0001/INDEX.HTM
   for useful notes 
   For complex numbers, http://www.clarku.edu/~djoyce/complex/
*)

let pi = 4.0 *. atan 1.0;;

type complex = Complex of (float * float)

let omega (n',n) =
  let z = -2.0 *. pi *. (float n) /. (float n') in 
    Complex (cos z , sin z)

let complex r = Complex (r,0.0);;

let (+..) a b =
  match (a,b) with
      (Complex (r,s), Complex (t,u)) -> 
	Complex (r+.t,s+.u);;

let (-..) a b =
  match (a,b) with
      (Complex (r,s), Complex (t,u)) -> 
	Complex (r-.t,s-.u);;

let ( *..) a b =
  match (a,b) with
      (Complex (x,y), Complex (u,v)) -> 
        Complex (x*.u -. y*.v, x*.v +. y*.u);;

let split n' f =
  (fun i -> f(i)+..(f(i+n'))),
  (fun i -> (f(i)-..(f(i+n')))*..omega(n',i))

let two fft n' (f,g) =
  (fft n' f),
  (fft n' g)

let merge (f,g) n =
  if (n mod 2) = 0
  then
    f (n / 2)
  else
    g ((n - 1) /2)

let rec fft n f =
if n=1 
then
  f
else
  let n' = n/2 in 
    merge (two fft n' (split n' f))

(* Staged *)

type dcomplex = DComplex of ((float code)* (float code))

let lift (Complex (x,y)) = DComplex (.<x>.,.<y>.)

let (+...) a b =
  match (a,b) with
      (DComplex (r,s), DComplex (t,u)) -> 
	DComplex (.< .~r+. .~t>., .< .~s+. .~u>.);;

let (-...) a b =
  match (a,b) with
      (DComplex (r,s), DComplex (t,u)) -> 
	DComplex (.< .~r-. .~t>., .<.~s-. .~u>.);;

let ( *...) a b =
  match (a,b) with
      (DComplex (x,y), DComplex (u,v)) -> 
        DComplex (.< .~x*. .~u -. .~y*. .~v>., 
                  .<.~x*. .~v +. .~y*. .~u>.);;

let split' n' f =
  (fun i -> f(i)+...(f(i+n'))),
  (fun i -> (f(i)-...(f(i+n')))*...(lift (omega(n',i))))

let rec fft' n f =
if n=1 
then
  f
else
  let n' = n/2 in 
    merge (two fft' n' (split' n' f));;

let u (DComplex (x,y)) = .<(.~x, .~y)>. ;;
let unfold f = .<(.~(u (f 0)),.~(u (f 1)),.~(u (f 2)),.~(u (f 3)))>. ;;
  
.<fun (x0,y0) ->
  fun (x1,y1) ->
  fun (x2,y2) ->
  fun (x3,y3) ->
.~(let f x =
     (match x with
	  0 -> DComplex (.<x0>., .<y0>.)
	| 1 -> DComplex (.<x1>., .<y1>.)
	| 2 -> DComplex (.<x2>., .<y2>.)
	| 3 -> DComplex (.<x3>., .<y3>.))
   in unfold (fft' 4 f))>.;;

(* Monadic version *)

let rec fftMinus (r,b) g (n,f) =
if n=1 
then
  r f
else
  let n' = n/2 in
  let (x,y) = split' n' f in
  b (g (n',x)) (fun v ->
  b (g (n',y)) (fun w ->
  r (merge (v,w))))

let m_ks  =
  let ret a k = k a in
  let bind a f k = a (fun x -> f x k) in
  (ret,bind);;

let ext f x y = fun z -> if z=x then y else f z

exception Undefined
let g0 x = raise Undefined

let rec name g n k g' =
if n = 0 then
  k g'
else
  let DComplex (x,y) = g (n-1) in
    .<let a = .~x in 
      let b = .~y in 
	.~(name g (n-1) k (ext g' (n-1) (DComplex (.<a>.,.<b>.))))>.

let rec y_sm f =
  f (fun (n,f') k ->
       y_sm f (n,f') (fun g -> name g n k g0))

let fft' (n,f) = y_sm (fftMinus m_ks) (n,f);;

.<fun (x0,y0) ->
  fun (x1,y1) ->
  fun (x2,y2) ->
  fun (x3,y3) ->
.~(let f x =
     (match x with
	  0 -> DComplex (.<x0>., .<y0>.)
	| 1 -> DComplex (.<x1>., .<y1>.)
	| 2 -> DComplex (.<x2>., .<y2>.)
	| 3 -> DComplex (.<x3>., .<y3>.))
   in (fft' (4,f) unfold))>.;;

(* a "known good" fft, although this rearranges and rotates some of its outputs (many FFTs do) *)
(*
#use "fft-baseline.ml";;
let known_good_fft n f =
  let rec populate a i = if i < 0 then a else (a.(i) <- (match (f i) with (x,y) -> (x,y)); (populate a (i-1))) in
  let a = FFT.fft (populate (Array.create(n) (0.0, 0.0)) (n-1)) in
  fun i -> a.(i)
let sample_known_good x = (float x,float x);;
let known_good = map (fun (x,y) -> (round x, round y)) (map (known_good_fft i sample0) (upto (i-1)));;
*)
